pub use crate::yl::get_1data_from_file;
pub use crate::yl::get_MM;
pub use crate::yl::get_data_from_file;
pub use crate::yl::get_dem;
pub use crate::yl::get_one_element;
pub use crate::yl::get_vec2_array_M;
pub use crate::yl::write_image_CM;
pub use crate::yl::write_image_arrow;

mod tecplot;

mod yl {
    use serde::{Deserialize, Serialize};
    use serde_json::json;
    use serde_json::{Result, Value};
    use std::env::Args;

    use regex::Regex;
    use std::fs::OpenOptions;
    use std::io::{Read, Write};
    use std::{
        fs::File,
        // io::Read,
        io::{BufRead, BufReader},
    };
    extern crate image;
    use image::ImageFormat;

    // extern crate num_complex;
    const UNUSED_OF_DEM: f64 = -9999.0;

    use cloudmap::cm3;
    use cloudmap::cm3_arrow;
    use cloudmap::cm3_blue;
    use cloudmap::cm3_mud;
    // use crate::FEA_code::get_dem;
    // use crate::FEA_code::get_data_from_file;
    // use crate::FEA_code::get_MM;
    // use crate::FEA_code::get_one_element_cms;

    pub fn get_dem(filename: &str) -> (Vec<f64>, i32, i32) {
        let mut cols = 10;
        let mut rows = 10;

        let file = File::open(filename).unwrap();
        let lines = BufReader::new(file).lines();
        //遍历所有行
        let mut i = 0;

        let mut list = Vec::new();
        let mut start_data = false;
        let seperator = Regex::new(r"([ ]+)").expect("Invalid regex");
        for line in lines {
            if let Ok(data) = line {
                let line_space: Vec<&str> = seperator.split(data.as_str()).into_iter().collect();
                if start_data == false {
                    if data.contains("ncols") {
                        // let line_space: Vec<&str> = seperator.split(data).into_iter().collect();
                        cols = line_space[1].parse::<i32>().unwrap();
                    } else if data.contains("nrows") {
                        // let line_space: Vec<&str> = seperator.split(data).into_iter().collect();
                        rows = line_space[1].parse::<i32>().unwrap();
                    } else if data.contains("NODATA_value") {
                        start_data = true;
                    }
                } else {
                    // let line_space: Vec<&str> = seperator.split(data).into_iter().collect();
                    if line_space.is_empty() {
                        start_data = false;
                        break;
                    }
                    let mut j = 0;
                    for one_str in line_space {
                        if one_str == "" {
                            break;
                        }
                        let value1: f64 = one_str.parse::<f64>().expect(
                            format!("dem file some vaule error,line= {},j={}", i, j).as_str(),
                        );
                        list.push(value1);
                        j += 1;
                    }
                }
            }
            i += 1;
        }
        (list, rows, cols)
    }

    pub fn get_1data_from_file(filename: &str, x: i32, y: i32) -> Vec<f64> {
        // Using BufReader because files in std is unbuffered by default
        // And reading by 8 bytes is really bad idea.
        let mut input = BufReader::new(File::open(filename).expect("Failed to open file"));

        // let mut floats: Vec<f64> = Vec::new();
        let mut zbed: Vec<f64> = Vec::new();

        let mut ij: u128 = 0;

        loop {
            use std::io::ErrorKind;

            let mut buffer = [0u8; std::mem::size_of::<f64>()];
            let res = input.read_exact(&mut buffer);
            match res {
                Err(error) if error.kind() == ErrorKind::UnexpectedEof => {
                    break;
                }
                _ => {ij += 1;}
            }
            res.expect("Unexpected error during read");

            // let f = f64::from_le_bytes(buffer);
            let z = f64::from_le_bytes(buffer);
            if ij==1 {
                print!("start :{:?}\n\n",z);
            }
            if ij==1591353 {
                print!("end :{:?}\n\n",z);
            }
            zbed.push(z);

            // ij += 1;
        }
        print!("data file :count={}, x:{x}*y:{y}={}\n", ij, x * y);
        // assert_eq!(
        //     ij,
        //     (x * y) as u128,
        //     "dem cols*row not equal *.dat cols*rows"
        // );

        
        zbed
    }

    pub fn get_data_from_file(filename: &str, x: i32, y: i32) -> (Vec<f64>, Vec<f64>, Vec<f64>) {
        // Using BufReader because files in std is unbuffered by default
        // And reading by 8 bytes is really bad idea.
        let mut input = BufReader::new(File::open(filename).expect("Failed to open file"));

        // let mut floats: Vec<f64> = Vec::new();
        let mut zbed: Vec<f64> = Vec::new();
        let mut uvu: Vec<f64> = Vec::new();
        let mut uvv: Vec<f64> = Vec::new();

        let mut ij: i32 = 0;

        loop {
            use std::io::ErrorKind;

            let mut buffer = [0u8; std::mem::size_of::<f64>()];
            let res = input.read_exact(&mut buffer);
            match res {
                Err(error) if error.kind() == ErrorKind::UnexpectedEof => {
                    break;
                }
                _ => {}
            }
            res.expect("Unexpected error during read");

            // let f = f64::from_le_bytes(buffer);
            let z = f64::from_le_bytes(buffer);

            buffer = [0u8; std::mem::size_of::<f64>()];
            let res = input.read_exact(&mut buffer);
            match res {
                Err(error) if error.kind() == ErrorKind::UnexpectedEof => {
                    break;
                }
                _ => {}
            }
            res.expect("Unexpected error during read");
            let u = f64::from_le_bytes(buffer);

            buffer = [0u8; std::mem::size_of::<f64>()];
            let res = input.read_exact(&mut buffer);
            match res {
                Err(error) if error.kind() == ErrorKind::UnexpectedEof => {
                    break;
                }
                _ => {}
            }
            res.expect("Unexpected error during read");
            let v = f64::from_le_bytes(buffer);

            zbed.push(z);
            uvu.push(u);
            uvv.push(v);

            ij += 1;
        }
        assert_eq!(ij, x * y, "dem cols*row not equal *.dat cols*rows");

        print!("data file :count={}, x*y={}\n", ij, x * y);
        (zbed, uvu, uvv)
    }

    pub fn get_MM(list: &Vec<f64>, dem: &Vec<f64>, x: i32, y: i32) -> (f64, f64) {
        let mut max = 0.;
        let mut min = 0.;

        let mut flag_first = false;

        let mut new_list: Vec<Vec<f64>> = Vec::new();

        let dem_len = dem.len() as i32;
        let i_dem = 1;
        let j_dem = 1;
        for i in 0..x {
            for j in 0..y {
                let pixel = *&dem[(i * (x) + (j)) as usize];
                let pixel_value = *&list[(i * (x + 2) + (j + 1)) as usize];

                if flag_first == false && pixel != -9999. {
                    max = pixel_value;
                    min = pixel_value;
                    flag_first = true;
                }
                if min > pixel_value {
                    min = pixel_value;
                }
                if max < pixel_value {
                    max = pixel_value;
                }
            }
        }

        (max, min)
    }

    pub fn get_one_element_cms(
        list: &Vec<f64>,
        dem: &Vec<f64>,
        max: f64,
        min: f64,
        x: i32,
        y: i32,
        cols: i32,
        rows: i32,
    ) -> (f64, f64, f64, f64, bool) {
        let dem_value_3 = *&dem[(y * cols + x) as usize];
        let mut el_value_3 = *&list[((y + 1) * (cols + 2) + x + 1) as usize];

        let dem_value_2 = *&dem[(y * cols + x + 1) as usize];
        let mut el_value_2 = *&list[((y + 1) * (cols + 2) + x + 1 + 1) as usize];

        let dem_value_0 = *&dem[(y * cols + x) as usize];
        let mut el_value_0 = *&list[((y + 2) * (cols + 2) + x + 1) as usize];

        let dem_value_1 = *&dem[(y * cols + x + 1) as usize];
        let mut el_value_1 = *&list[((y + 2) * (cols + 2) + x + 1 + 1) as usize];

        let mut flag = 1;

        if dem_value_3 != UNUSED_OF_DEM {
            el_value_3 = (el_value_3 - min) / (max - min);
        } else {
            //el_value_3 = (el_value_3 - min) / (max - min);
            return (-9., -9., -9., -9., false);
        }

        if dem_value_0 != UNUSED_OF_DEM {
            el_value_0 = (el_value_0 - min) / (max - min);
        } else {
            el_value_0 = el_value_3 / 2.0;
            return (-9., -9., -9., -9., false);
        }

        if dem_value_1 != UNUSED_OF_DEM {
            el_value_1 = (el_value_1 - min) / (max - min);
        } else {
            el_value_1 = (el_value_3) / 2.0;
            return (-9., -9., -9., -9., false);
        }

        if dem_value_2 != UNUSED_OF_DEM {
            el_value_2 = (el_value_2 - min) / (max - min);
        } else {
            el_value_2 = (el_value_3) / 3.0;
            return (-9., -9., -9., -9., false);
        }

        // if el_value_0 == 1. || el_value_1 == 1. || el_value_2 == 1. || el_value_3 == 1. {
        //     print!("hkjhk");
        // }
        if (x == 313) && (y == 138) {
            // return (1.0, 1.0, 1.0, 1.0, true);
        }
        return (el_value_0, el_value_1, el_value_2, el_value_3, true);
    }

    pub fn get_one_element(
        list: &Vec<f64>,
        dem: &Vec<f64>,
        max: f64,
        min: f64,
        x: i32,
        y: i32,
        cols: i32,
        rows: i32,
    ) -> f64 {
        let dem_value = *&dem[(y * cols + x) as usize];
        let mut el_value = *&list[((y + 1) * (cols + 2) + x + 1) as usize];

        if dem_value == -9999. {
            return -9999.;
        } else {
            return (el_value - min) / (max - min);
        }
    }

    pub fn get_one_element_2uv(
        u: &Vec<f64>,
        v: &Vec<f64>,
        dem: &Vec<f64>,
        max: f64,
        min: f64,
        x: i32,
        y: i32,
        cols: i32,
        rows: i32,
    ) -> (f64, f64) {
        let dem_value = *&dem[(y * cols + x) as usize];
        let el_u = *&u[(y * (cols + 2) + x + 1) as usize];
        let el_v = *&v[(y * (cols + 2) + x + 1) as usize];

        if dem_value == -9999. {
            return (-9999., -9999.);
        } else {
            // print!(
            //     "max:{max},min:{min},U:{el_u},V:{el_v},{},{}\n",
            //     (el_u - min) / (max - min),
            //     (el_v - min) / (max - min)
            // );
            return (
                ((el_u - min) / (max - min)) * 2.0 - 1.0,
                ((el_v - min) / (max - min)) * 2.0 - 1.0,
            );
        }
    }

    pub fn get_vec2_array_M(list_u: &Vec<f64>, list_v: &Vec<f64>) -> Vec<f64> {
        let u = list_u.clone();
        let v = list_v.clone();
        let mut M = Vec::new();
        let mut ii: usize = 0;
        let len = u.len() as usize;
        while ii < len {
            let i = *u.get(ii).unwrap();
            let j = *v.get(ii).unwrap();
            let m = i * i + j * j;
            ii += 1;
            M.push(m.sqrt());
        }
        M
    }

    pub fn write_image_CM(
        list: &Vec<f64>,
        dem: &Vec<f64>,
        max: f64,
        min: f64,
        cols: i32,
        rows: i32,
        pn: i32,
        cut: f64,
        filename: &str,
        color: &str,
    ) {
        let imgx = (cols * pn) as u32;
        let imgy = (rows * pn) as u32;

        // Create a new ImgBuf with width: imgx and height: imgy
        let mut imgbuf = image::ImageBuffer::new(imgx, imgy);

        for x in 0..cols - 1 {
            //  print!("行：   {}\n", x);
            for y in 0..rows - 1 {
                let (ca, cb, cc, cd, flag) =
                    get_one_element_cms(list, dem, max, min, x, y, cols, rows);
                if flag {
                    for rx in 0..pn {
                        for ry in 00..pn {
                            let pixel =
                                imgbuf.get_pixel_mut((x * pn + rx) as u32, (y * pn + ry) as u32);
                            let image::Rgba(data) = *pixel;
                            let mut cmv: [u32; 4];
                            if ry >= pn - rx {
                                if color == "blue" {
                                    cmv = cm3_blue(
                                        rx as f64 / pn as f64,
                                        1.0 - (ry as f64 / pn as f64),
                                        ca,
                                        cb,
                                        cc,
                                        cut,
                                        12,
                                    );
                                } else {
                                    cmv = cm3(
                                        rx as f64 / pn as f64,
                                        1.0 - (ry as f64 / pn as f64),
                                        ca,
                                        cb,
                                        cc,
                                        cut,
                                        12,
                                    );
                                }
                            } else {
                                if color == "blue" {
                                    cmv = cm3_blue(
                                        (pn - ry) as f64 / pn as f64,
                                        (rx) as f64 / pn as f64,
                                        ca,
                                        cd,
                                        cc,
                                        cut,
                                        12,
                                    );
                                } else {
                                    cmv = cm3(
                                        (pn - ry) as f64 / pn as f64,
                                        (rx) as f64 / pn as f64,
                                        ca,
                                        cd,
                                        cc,
                                        cut,
                                        12,
                                    );
                                }
                            }

                            *pixel = image::Rgba([
                                cmv[0] as u8,
                                cmv[1] as u8,
                                cmv[2] as u8,
                                cmv[3] as u8,
                            ]);
                        }
                    }
                }
            }
        }
        print!("ready to  write cm\n");
        // Save the image as “fractal.png”, the format is deduced from the path
        imgbuf.save(filename).unwrap();
        print!("end write cm\n");
        // imgbuf.save_with_format("cm.Jpeg",ImageFormat::Jpeg).unwrap();
    }

    pub fn write_image_arrow(
        list_u: &Vec<f64>,
        list_v: &Vec<f64>,
        dem: &Vec<f64>,
        max: f64,
        min: f64,
        // u_max: f64,
        // u_min: f64,
        // v_max: f64,
        // v_min: f64,
        cols: i32,
        rows: i32,
        pn: i32,
        filename: &str,
        filter_list: &Vec<f64>,
        filter_max: f64,
        filter_min: f64,
        cut: f64,
    ) {
        let imgx = (cols * pn) as u32;
        let imgy = (rows * pn) as u32;

        // Create a new ImgBuf with width: imgx and height: imgy
        let mut imgbuf = image::ImageBuffer::new(imgx, imgy);

        for x in 0..cols - 1 {
            //  print!("行：   {}\n", x);
            for y in 0..rows - 1 {
                let (ux, vy) = get_one_element_2uv(list_u, list_v, dem, max, min, x, y, cols, rows);
                // let ux =
                //     get_one_element(list_u, dem, u_max, u_min, x, y, cols, rows);
                // let vy =
                //     get_one_element(list_v, dem, v_max, v_min, x, y, cols, rows);
                let vv =
                    get_one_element(filter_list, dem, filter_max, filter_min, x, y, cols, rows);
                if vv > cut {
                    let abc = 1;
                    for rx in 0..pn {
                        for ry in 00..pn {
                            let pixel =
                                imgbuf.get_pixel_mut((x * pn + rx) as u32, (y * pn + ry) as u32);
                            let image::Rgba(data) = *pixel;
                            let mut cmv: [u32; 4];

                            let mut u = 0.0;
                            let mut v = 0.0;

                            if pn - ry <= rx {
                                u = ((rx as f64) * 2.0 - (pn as f64)) / pn as f64;
                                v = ((ry as f64) * 2.0 - (pn as f64)) / pn as f64;
                                cmv = cm3_arrow(ux, vy, u, v, vv, cut, 12);
                            } else {
                                u = ((ry as f64) * 2.0 - (pn as f64)) / pn as f64;
                                v = ((rx as f64) * 2.0 - (pn as f64)) / pn as f64;
                                cmv = cm3_arrow(ux, vy, v, u, vv, cut, 12);
                            }

                            *pixel = image::Rgba([
                                cmv[0] as u8,
                                cmv[1] as u8,
                                cmv[2] as u8,
                                cmv[3] as u8,
                            ]);
                        }
                    }
                }
            }
        }

        // Save the image as “fractal.png”, the format is deduced from the path
        imgbuf.save(filename).unwrap();
        // imgbuf.save_with_format("cm.Jpeg",ImageFormat::Jpeg).unwrap();
    }
}
